from __future__ import annotations

from typing import TYPE_CHECKING, Final

from sml2mqtt.runtime import do_shutdown

from .device_status import DeviceStatus


if TYPE_CHECKING:
    from . import SmlDevice


class SmlDevices:
    def __init__(self) -> None:
        self._devices: tuple[SmlDevice, ...] = ()

    def add_device(self, device: SmlDevice) -> SmlDevice:
        for existing in self._devices:
            if existing.name == device.name:
                msg = f'Device {device.name:s} does already exist!'
                raise ValueError(msg)
        self._devices = (*self._devices, device)
        return device

    async def start(self) -> None:
        for device in self._devices:
            await device.start()

    async def cancel_and_wait(self) -> None:
        for device in self._devices:
            await device.cancel_and_wait()

    def check_status(self):
        if any(device.status in (DeviceStatus.SOURCE_FAILED, DeviceStatus.SHUTDOWN) for device in self._devices):
            return do_shutdown()

        if all(device.status.is_shutdown_status() for device in self._devices):
            return do_shutdown()

        return None

    def __len__(self) -> int:
        return len(self._devices)


ALL_DEVICES: Final = SmlDevices()



from __future__ import annotations

from datetime import date, time, timedelta
from enum import Enum
from typing import Annotated, Final, Literal, TypeAlias, TypedDict, Union, final
from typing import get_args as _get_args

from annotated_types import Len
from easyconfig import BaseModel
from pydantic import Discriminator, Field, StrictBool, StrictFloat, StrictInt, Tag, model_validator

from sml2mqtt.const import DateTimeFinder, DurationType, TimeSeries

from .types import Number, ObisHex  # noqa: TCH001


class EmptyKwargs(TypedDict):
    pass


# -------------------------------------------------------------------------------------------------
# Filters
# -------------------------------------------------------------------------------------------------
class OnChangeFilter(BaseModel):
    """A filter which lets the value only pass when it's different from the value that was passed the last time"""
    type: Literal['change filter'] = Field(description='Filter which passes only changes')

    @final
    def get_kwargs_on_change(self) -> EmptyKwargs:
        return {}


class RangeFilter(BaseModel):
    """Filters or limits to values that are in a certain range

    """
    type: Literal['range filter']
    min_value: float | None = Field(None, alias='min', description='minimum value that will pass')
    max_value: float | None = Field(None, alias='max', description='maximum value that will pass')
    limit_values: bool = Field(
        False, alias='limit', description='Instead of ignoring the values they will be limited to min/max'
    )

    @model_validator(mode='after')
    def _check_set(self) -> RangeFilter:
        if self.min_value is None and self.max_value is None:
            msg = 'Neither min or max are set!'
            raise ValueError(msg)
        return self


class DeltaFilter(BaseModel):
    """A filter which lets the value only pass if the incoming value is different enough from value that was passed the
    last time. The delta can an absolute value or as a percentage.
    If multiple deltas are specified they are all checked.
    """

    type: Literal['delta filter']

    min_value: StrictInt | StrictFloat | None = Field(None, alias='min')
    min_percent: StrictInt | StrictFloat | None = Field(None, alias='min %')

    @model_validator(mode='after')
    def _check_set(self) -> DeltaFilter:
        if self.min_value is None and self.min_percent is None:
            msg = 'Neither min or min % are set!'
            raise ValueError(msg)
        return self


class ThrottleFilter(BaseModel):
    """Filter which only lets one value pass in the defined period. If the last passed value is not at least
    ``period`` old any new value will not be forwarded.
    """
    period: DurationType = Field(alias='throttle filter', description='Throttle period')


# -------------------------------------------------------------------------------------------------
# Actions
# -------------------------------------------------------------------------------------------------
class RefreshAction(BaseModel):
    """Action which lets every value pass. When no value is received (e.g. because an earlier filter blocks)
    this action will produce the last received value every interval.
    """
    every: DurationType = Field(alias='refresh action', description='Refresh interval')


class HeartbeatAction(BaseModel):
    """Action which lets a value pass periodically every specified interval.
    When no value is received (e.g. because an earlier filter blocks)
    this action will produce the last received value every interval.
    """

    every: DurationType = Field(
        alias='heartbeat action',
        description='Interval'
    )


# -------------------------------------------------------------------------------------------------
# Math
# -------------------------------------------------------------------------------------------------
class Factor(BaseModel):
    factor: Number = Field(description='Factor with which the value gets multiplied')


class Offset(BaseModel):
    offset: Number = Field(description='Offset that gets added on the value')


class Round(BaseModel):
    digits: int = Field(ge=-4, le=6, alias='round', description='Round to the specified digits, negative for tens')


class RoundToMultiple(BaseModel):
    """Rounds to the multiple of a given value.
    """

    type: Literal['round to multiple']
    value: int = Field(description='Round to the multiple of the given value')
    round: Literal['up', 'down', 'nearest'] = Field(description='Round up, down or to the nearset multiple')


# -------------------------------------------------------------------------------------------------
# Workarounds
# -------------------------------------------------------------------------------------------------
class NegativeOnEnergyMeterWorkaround(BaseModel):
    """Make value negative based on an energy meter status."""

    enabled_or_obis: StrictBool | ObisHex = Field(
        alias='negative on energy meter status',
        description='Set to "true" to enable or to "false" to disable workaround. '
                    'If the default obis code for the energy meter is wrong set '
                    'to the appropriate meter obis code instead'
    )


# -------------------------------------------------------------------------------------------------
# Operations
# -------------------------------------------------------------------------------------------------
class Or(BaseModel):
    """A sequence of operations that will be evaluated one after another. The first value that gets returned by an
    operation will be used.
    """
    operations: OperationsListType = Field(alias='or')


class Sequence(BaseModel):
    """A sequence of operations that will be evaluated one after another.
    If one operation blocks this will return nothing.
    """
    operations: OperationsListType = Field(alias='sequence')


def generate_day_names() -> dict[str, int]:
    # names of weekdays in local language
    day_names: dict[str, int] = {date(2001, 1, i).strftime('%A'): i for i in range(1, 8)}
    day_names.update({date(2001, 1, i).strftime('%A')[:3]: i for i in range(1, 8)})

    # abbreviations in German and English
    day_names.update({'Mo': 1, 'Di': 2, 'Mi': 3, 'Do': 4, 'Fr': 5, 'Sa': 6, 'So': 7})
    day_names.update({'Mon': 1, 'Tue': 2, 'Wed': 3, 'Thu': 4, 'Fri': 5, 'Sat': 6, 'Sun': 7})
    return {k.lower(): v for k, v in day_names.items()}


DayOfWeekStr = Enum('DayOfWeekStr', {k: k for k in generate_day_names()}, type=str)
DayOfMonth = Annotated[int, Field(ge=1, le=31, strict=True)]


# -------------------------------------------------------------------------------------------------
# DateTime
# -------------------------------------------------------------------------------------------------
class HasDateTimeFields(BaseModel):

    start_now: bool = Field(
        alias='start now', description='Immediately start instead of starting after the next reset'
    )
    reset_times: list[time] = Field(
        default=[], alias='reset times', description='Time(s) of day when a reset will occur',
    )
    reset_days: list[DayOfMonth | DayOfWeekStr] = Field(
        default=[], alias='reset days', description='Days of month or weekdays where the time(s) will be checked'
    )

    @final
    def get_kwargs_dt_fields(self) -> DateTimeBoundKwargs:

        names = generate_day_names()
        dows = [names[n] for n in self.reset_days if not isinstance(n, int)]
        days = [n for n in self.reset_days if isinstance(n, int)]

        finder = DateTimeFinder()
        for t in self.reset_times:
            finder.add_time(t)
        for dow in dows:
            finder.add_dow(dow)
        for day in days:
            finder.add_day(day)

        return {
            'dt_finder': finder,
            'start_now': self.start_now
        }


class DateTimeBoundKwargs(TypedDict):
    start_now: bool
    dt_finder: DateTimeFinder


class VirtualMeter(HasDateTimeFields):
    """A virtual meter. It will output the difference from the last reset"""
    type: Literal['meter']


class MaxValue(HasDateTimeFields):
    """Maximum value since last reset"""
    type: Literal['max value']


class MinValue(HasDateTimeFields):
    """Minimum value since last reset"""
    type: Literal['min value']


# -------------------------------------------------------------------------------------------------
# TimeSeries
# -------------------------------------------------------------------------------------------------
class TimeSeriesKwargs(TypedDict):
    time_series: TimeSeries
    reset_after_value: bool


class HasIntervalFields(BaseModel):

    interval: timedelta = Field(
        description='Interval duration'
    )

    wait_for_data: bool = Field(
        alias='wait for data', description='Only produce a value when data for the whole interval is available'
    )

    reset_after_value: bool = Field(
        False, alias='reset after value', description='Clear all data as soon as a value has been produced'
    )

    @final
    def get_kwargs_interval_fields(self) -> TimeSeriesKwargs:
        return {
            'time_series': TimeSeries(self.interval, wait_for_data=self.wait_for_data),
            'reset_after_value': self.reset_after_value
        }


class MaxOfInterval(HasIntervalFields):
    """Maximum value in a sliding interval"""
    type: Literal['max interval']


class MinOfInterval(HasIntervalFields):
    """Minimum value in a sliding interval"""
    type: Literal['min interval']


class MeanOfInterval(HasIntervalFields):
    """Weighted mean in a sliding interval"""
    type: Literal['mean interval']


# -------------------------------------------------------------------------------------------------

OperationsModels = (
    OnChangeFilter, DeltaFilter, HeartbeatAction, RangeFilter,
    RefreshAction, ThrottleFilter,
    Factor, Offset, Round, RoundToMultiple,
    NegativeOnEnergyMeterWorkaround,
    Or, Sequence,
    VirtualMeter, MaxValue, MinValue,
    MaxOfInterval, MinOfInterval, MeanOfInterval,
)

# noinspection PyTypeHints
OperationsType: TypeAlias = Union[tuple(Annotated[o, Tag(o.__name__)] for o in OperationsModels)]  # noqa: UP007


MODEL_FIELD_MAP: Final[dict[str, frozenset[str]]] = {
    _m.__name__: frozenset(
        _f.alias if _f.alias is not None else _n for _n, _f in _m.model_fields.items() if _f.exclude is not True
    )
    for _m in OperationsModels
}

MODEL_TYPE_MAP: Final[dict[str, str]] = {
    _get_args(_f.annotation)[0]: _m.__name__
    for _m in OperationsModels for _n, _f in _m.model_fields.items() if _n == 'type'
}


def check_allowed_keys(obj: dict | BaseModel) -> str | None:
    if isinstance(obj, dict):
        type: str = obj.get('type')  # noqa: A001
        keys: set[str] = set(obj)
    else:
        type: str = getattr(obj, 'type', None)  # noqa: A001
        keys: set[str] = set(obj.__class__.model_fields)

    # we have a type field
    if type is not None:
        return MODEL_TYPE_MAP.get(type)

    # let's see if we have a 100% match
    for name, fields in MODEL_FIELD_MAP.items():
        if keys == fields:
            return name
    return None


OperationsTypeAnnotated: TypeAlias = Annotated[
    OperationsType,
    Discriminator(
        check_allowed_keys,
        custom_error_type='invalid_key_names',
        custom_error_message='Invalid key names',
        custom_error_context={'discriminator': 'check_allowed_keys'}
    )
]

OperationsListType = Annotated[list[OperationsTypeAnnotated], Len(min_length=0)]


def cleanup_validation_errors(msg: str) -> str:
    # In the ValidationError there is the Model and the field, but the user should only be concerned by the field name
    return msg.replace('Or.or', 'or').replace('Sequence.sequence', 'sequence')



from datetime import datetime, timedelta
from datetime import time as dt_time


def get_now():
    return datetime.now()


class DateTimeFinder:
    def __init__(self) -> None:
        self.times: tuple[dt_time, ...] = ()
        self.dows: tuple[int, ...] = ()
        self.days: tuple[int, ...] = ()

        self.enabled: bool = False

    @property
    def condition_count(self) -> int:
        return len(self.times) + len(self.dows) + len(self.days)

    def add_time(self, time: dt_time):
        if not isinstance(time, dt_time):
            raise TypeError()
        self.times = (*self.times, time)
        self.enabled = True
        return self

    def add_dow(self, dow: int):
        if not isinstance(dow, int):
            raise TypeError()
        if dow < 1 or dow > 7 or dow in self.dows:
            raise ValueError()
        self.dows = (*self.dows, dow)
        return self

    def add_day(self, day: int):
        if not isinstance(day, int):
            raise TypeError()
        if day < 1 or day > 31 or day in self.days:
            raise ValueError()
        self.days = (*self.days, day)
        return self

    def get_first_reset(self, start_now: bool):
        return get_now() if start_now or not self.enabled else self.calc_next()

    def calc_next(self, now: datetime | None = None) -> datetime:
        if now is None:
            now = get_now()

        next_dt = now
        while True:
            date = next_dt.date()
            for time in self.times:
                if (new := datetime.combine(date, time)) > now:
                    return new

            while True:
                next_dt += timedelta(days=1)
                if not self.dows and not self.days or next_dt.isoweekday() in self.dows or next_dt.day in self.days:
                    break



import re

from easyconfig.yaml import yaml_rt

from sml2mqtt.config.config import Settings


def test_readme(pytestconfig) -> None:

    readme = pytestconfig.rootpath / 'readme.md'
    assert readme.is_file()

    yaml_parts = re.findall(r'```ya?ml\n(.+?)```', readme.read_text(encoding='utf-8'), re.DOTALL)

    # First entry is the complete config
    for cfg_sample in yaml_parts:
        obj = yaml_rt.load(cfg_sample)    # type: dict
        Settings(**obj)




def format_period(period: int | float) -> str:
    period = period
    h = int(period) // 3600
    period %= 3600
    m = int(period) // 60
    s = period % 60

    parts = []
    for part, unit in ((h, 'hour'), (m, 'minute'), (s, 'second')):
        if not part:
            continue
        if part != 1:
            unit += 's'
        parts.append(f'{part:d} {unit:s}' if isinstance(part, int) else f'{part:.1f} {unit:s}')
    return ' '.join(parts)



from asyncio import sleep
from types import TracebackType


class DynDelay:
    def __init__(self, min_delay: float, max_delay: float, start_delay: float | None = None) -> None:
        if min_delay < 0:
            msg = f'min_delay must be >= 0: {min_delay}'
            raise ValueError(msg)
        if max_delay <= min_delay:
            msg = f'max_delay must be >= min_delay: {max_delay}'
            raise ValueError(msg)

        self.min = min_delay
        self.max = max_delay

        if start_delay is None:
            start_delay = min_delay

        self.curr = max(min(start_delay, max_delay), min_delay)

    async def wait(self) -> None:
        await sleep(self.curr)

    def increase(self) -> None:
        self.curr = min(self.max, self.curr * 2)
        if not self.curr:
            self.curr = 1

    def reset(self) -> None:
        self.curr = self.min

    async def __aenter__(self) -> None:
        await self.wait()

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> bool:  # noqa: E501
        if exc_type is None:
            self.reset()
        else:
            self.increase()
        return False



import traceback
from asyncio import CancelledError, Event, Queue, TimeoutError, wait_for
from typing import Final

from aiomqtt import Client, MqttError, Will

import sml2mqtt
from sml2mqtt.__log__ import log as _parent_logger
from sml2mqtt.const import Task
from sml2mqtt.errors import InitialMqttConnectionFailedError
from sml2mqtt.mqtt import DynDelay
from sml2mqtt.runtime import on_shutdown


log = _parent_logger.getChild('mqtt')


TASK: Task | None = None
IS_CONNECTED: Event | None = None


async def start() -> None:
    global IS_CONNECTED, TASK

    assert TASK is None

    IS_CONNECTED = Event()
    TASK = Task(_mqtt_task, name='MQTT Task')

    on_shutdown(TASK.cancel_and_wait, 'Shutdown mqtt')
    TASK.start()


async def wait_for_connect(timeout: float):
    if IS_CONNECTED is None:
        raise ValueError()

    try:
        await wait_for(IS_CONNECTED.wait(), timeout)
    except TimeoutError:
        log.error('Initial mqtt connection failed!')
        raise InitialMqttConnectionFailedError() from None

    return None


QUEUE: Queue[tuple[str, int | float | str | bytes, int, bool]] | None = None


async def _mqtt_task() -> None:
    global QUEUE

    from .mqtt_obj import BASE_TOPIC
    config = sml2mqtt.config.CONFIG

    cfg_connection = config.mqtt.connection

    delay = DynDelay(0, 300)

    payload_online: Final = 'ONLINE'
    payload_offline: Final = 'OFFLINE'

    shutdown = False

    while not shutdown:
        await delay.wait()

        try:
            # since we just pass this into the mqtt wrapper we do not link it to the base topic
            will_topic = BASE_TOPIC.create_child(topic_fragment='status').set_config(config.mqtt.last_will)

            will = Will(
                topic=will_topic.topic, payload=payload_offline,
                qos=will_topic.qos, retain=will_topic.retain
            )

            tls_kwargs = {} if cfg_connection.tls is None else cfg_connection.tls.get_client_kwargs(log)

            client = Client(
                hostname=cfg_connection.host, port=cfg_connection.port,

                username=cfg_connection.user if cfg_connection.user else None,
                password=cfg_connection.password if cfg_connection.password else None,
                will=will,
                identifier=cfg_connection.identifier,

                **tls_kwargs
            )

            log.debug(f'Connecting to {cfg_connection.host}:{cfg_connection.port}')

            async with client:
                log.debug('Success!')
                delay.reset()
                QUEUE = Queue()
                IS_CONNECTED.set()

                try:
                    # signal that we are online
                    await client.publish(will.topic, payload_online, will.qos, will.retain)

                    # worker to publish things
                    while True:
                        topic, value, qos, retain = await QUEUE.get()
                        await client.publish(topic, value, qos, retain)
                        QUEUE.task_done()

                except CancelledError:
                    # The last will testament only gets sent on abnormal disconnect
                    # Since we disconnect gracefully we have to manually sent the offline status
                    await client.publish(will.topic, will.payload, will.qos, will.retain)
                    shutdown = True
                    log.debug('Disconnecting')

        except MqttError as e:
            delay.increase()
            log.error(f'{e} ({e.__class__.__name__})')

        except Exception:
            delay.increase()
            for line in traceback.format_exc().splitlines():
                log.error(line)

        finally:
            QUEUE = None
            IS_CONNECTED.clear()


def publish(topic: str, value: int | float | str | bytes, qos: int, retain: bool) -> None:
    if QUEUE is not None:
        QUEUE.put_nowait((topic, value, qos, retain))



from __future__ import annotations

from asyncio import TimeoutError, sleep
from typing import TYPE_CHECKING, Final

from aiohttp import BasicAuth, ClientError, ClientSession, ClientTimeout

from sml2mqtt.__log__ import get_logger
from sml2mqtt.const import DeviceTask
from sml2mqtt.errors import HttpStatusError
from sml2mqtt.runtime import on_shutdown


if TYPE_CHECKING:
    from sml2mqtt.config.inputs import HttpSourceSettings
    from sml2mqtt.const import DeviceProto


log = get_logger('http')

SESSION: ClientSession | None = None


async def get_session() -> ClientSession:
    global SESSION

    if SESSION is not None:
        return SESSION

    SESSION = ClientSession()
    on_shutdown(close_session, 'Close http session')
    return SESSION


async def close_session():
    global SESSION

    if (session := SESSION) is None:
        return None

    SESSION = None
    await session.close()

    # https://docs.aiohttp.org/en/stable/client_advanced.html#graceful-shutdown
    await sleep(0.250)


class HttpSource:

    @classmethod
    async def create(cls, device: DeviceProto, settings: HttpSourceSettings):
        auth = None
        if settings.user or settings.password:
            auth = BasicAuth(settings.user, settings.password)

        return cls(device, str(settings.url), settings.interval, auth, timeout=settings.get_request_timeout())

    def __init__(self, device: DeviceProto,
                 url: str, interval: float,
                 auth: BasicAuth | None, timeout: ClientTimeout) -> None:
        super().__init__()
        self.device: Final = device

        self.url: Final = url
        self.auth: Final = auth
        self.timeout: Final = timeout

        self.interval = interval
        self._task: Final = DeviceTask(device, self._http_task, name=f'Http Task {self.device.name:s}')

    def start(self) -> None:
        self._task.start()

    async def cancel_and_wait(self) -> bool:
        return await self._task.cancel_and_wait()

    async def _http_task(self) -> None:
        log.debug(f'Requesting data from {self.url}')

        try:
            session = await get_session()
        except Exception as e:
            self.device.on_source_failed(f'Could not create client session: {e}')
            return None

        interval: float = self.interval
        com_errors: int = 0

        while True:
            await sleep(interval)
            interval = self.interval

            try:
                resp = await session.get(self.url, auth=self.auth, timeout=self.timeout)
                if resp.status != 200:
                    raise HttpStatusError(resp.status)  # noqa: TRY301

                payload = await resp.read()
                com_errors = 0
            except Exception as e:
                if isinstance(e, (ClientError, HttpStatusError, TimeoutError)):
                    com_errors += 1
                    max_ignore: int = 7
                    if com_errors <= max_ignore:
                        # errors:   1,   2,   3,   4,   5,   6,    7,    8,    9, ...
                        # factor: 0.4, 0.9, 1.9, 3.4, 5.4, 7.9, 10.9, 14.4, 18.4, ...
                        interval = (((com_errors - 0.5) ** 2) / 4 + 0.3) * self.interval
                        log.debug(f'Ignored {com_errors:d}/{max_ignore:d} {e} ({type(e)})')
                        continue

                self.device.on_error(e, show_traceback=False)
                continue

            self.device.on_source_data(payload)



from typing import Annotated, Literal, TypeAlias

import serial
from aiohttp import ClientTimeout
from easyconfig import BaseModel, Field
from pydantic import (
    AnyHttpUrl,
    StrictFloat,
    StrictInt,
    constr,
    field_validator,
    model_validator,
)
from typing_extensions import override

from sml2mqtt.config.types import log


class SmlSourceSettingsBase(BaseModel):
    def get_device_name(self) -> str:
        raise NotImplementedError()


class SerialSourceSettings(SmlSourceSettingsBase):
    type: Literal['serial']

    url: constr(strip_whitespace=True, min_length=1, strict=True) = Field(..., description='Device path')
    timeout: StrictInt | StrictFloat = Field(
        default=6, description='Seconds after which a timeout will be detected (default=6)')

    baudrate: int = Field(9600, in_file=False)
    parity: str = Field('None', in_file=False)
    stopbits: StrictInt | StrictFloat = Field(serial.STOPBITS_ONE, in_file=False, alias='stop bits')
    bytesize: int = Field(serial.EIGHTBITS, in_file=False, alias='byte size')

    @field_validator('baudrate')
    def _val_baudrate(cls, v):
        if v not in serial.Serial.BAUDRATES:
            msg = f'must be one of {list(serial.Serial.BAUDRATES)}'
            raise ValueError(msg)
        return v

    @field_validator('parity')
    def _val_parity(cls, v):
        # Short name
        if v in serial.PARITY_NAMES:
            return v

        # Name -> Short name
        parity_values = {_n: _v for _v, _n in serial.PARITY_NAMES.items()}
        if v not in parity_values:
            msg = f'must be one of {list(parity_values)}'
            raise ValueError(msg)
        return parity_values[v]

    @field_validator('stopbits')
    def _val_stopbits(cls, v):
        if v not in serial.Serial.STOPBITS:
            msg = f'must be one of {list(serial.Serial.STOPBITS)}'
            raise ValueError(msg)
        return v

    @field_validator('bytesize')
    def _val_bytesize(cls, v):
        if v not in serial.Serial.BYTESIZES:
            msg = f'must be one of {list(serial.Serial.BYTESIZES)}'
            raise ValueError(msg)
        return v

    @override
    def get_device_name(self) -> str:
        return self.url.split('/')[-1]


class HttpSourceSettings(SmlSourceSettingsBase):
    type: Literal['http']

    url: AnyHttpUrl = Field(..., description='Url')
    timeout: StrictInt | StrictFloat = Field(
        default=6, description='Seconds after which a timeout will be detected (default=6)')

    interval: StrictInt | StrictFloat = Field(default=2, description='Delay between requests', ge=0.1)
    user: str = Field(default='', description='User (if needed)')
    password: str = Field(default='', description='Password (if needed)')

    request_timeout: StrictInt | StrictFloat | None = Field(
        default=None, alias='request timeout', description='Dedicated timeout for the http request',
        in_file=False
    )

    @override
    def get_device_name(self) -> str:
        return self.url.host

    @model_validator(mode='after')
    def check_timeout_gt_interval(self):
        if self.interval * 2 > self.timeout:
            msg = 'Timeout must be greater equal than 2 * interval'
            raise ValueError(msg)

        # Timeout is interval, and we automatically retry 3 times before we fail
        if self.interval * 3 > self.timeout:
            log.warning(f'The recommendation for timeout is at least 3 * interval '
                        f'({self.interval * 3:.0f})! Currently: {self.timeout}')

        return self

    def get_request_timeout(self) -> ClientTimeout:
        value = self.interval if self.request_timeout is None else self.request_timeout
        if value is None:
            msg = 'No value for ClientTimeout'
            raise ValueError(msg)
        return ClientTimeout(total=value)


SmlSourceSettingType: TypeAlias = Annotated[HttpSourceSettings | SerialSourceSettings, Field(discriminator='type')]



import asyncio
import os
import sys
import traceback

from sml2mqtt import mqtt
from sml2mqtt.__args__ import CMD_ARGS, get_command_line_args
from sml2mqtt.__log__ import log, setup_log
from sml2mqtt.config import CONFIG, cleanup_validation_errors
from sml2mqtt.const.task import wait_for_tasks
from sml2mqtt.runtime import do_shutdown_async, on_shutdown, signal_handler_setup
from sml2mqtt.sml_device import ALL_DEVICES, SmlDevice
from sml2mqtt.sml_source import create_source


async def a_main() -> None:
    # Add possibility to stop program with Ctrl + c
    signal_handler_setup()

    on_shutdown(ALL_DEVICES.cancel_and_wait, 'Stop devices')

    try:
        if analyze := CMD_ARGS.analyze:
            mqtt.patch_analyze()
        else:
            # initial mqtt connect
            await mqtt.start()
            await mqtt.wait_for_connect(5)

        # Create device for each input
        for input_cfg in CONFIG.inputs:
            device = ALL_DEVICES.add_device(SmlDevice(input_cfg.get_device_name()))
            device.set_source(await create_source(device, settings=input_cfg))
            device.watchdog.set_timeout(input_cfg.timeout)
            if analyze:
                device.frame_handler = device.analyze_frame

        # Start all devices
        log.debug(f'Starting {len(ALL_DEVICES):d} device{"" if len(ALL_DEVICES) == 1 else "s":s}')
        await ALL_DEVICES.start()

    except Exception as e:
        log.error(f'{e.__class__.__name__} during startup: {e}')
        for line in traceback.format_exc().splitlines():
            log.error(line)

        await do_shutdown_async()

    # Keep tasks running
    await wait_for_tasks()


def main() -> int | str:
    # This is needed to make async-mqtt work
    # see https://github.com/sbtinstruments/asyncio-mqtt
    if sys.platform.lower() == 'win32' or os.name.lower() == 'nt':
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    # Load config
    try:
        CONFIG.load_config_file(get_command_line_args().config)
    except Exception as e:
        print(cleanup_validation_errors(str(e)))
        return 2

    try:
        setup_log()

        # setup mqtt base topic
        mqtt.setup_base_topic(CONFIG.mqtt.topic, CONFIG.mqtt.defaults.qos, CONFIG.mqtt.defaults.retain)

        asyncio.run(a_main())
    except Exception as e:
        for line in traceback.format_exc().splitlines():
            log.error(line)
            print(e)
        return str(e)

    return 0


if __name__ == '__main__':
    ret = main()
    log.info(f'Closed with return code {ret}')
    sys.exit(ret)



from __future__ import annotations

from smllib import SmlFrame, SmlStreamReader
from smllib.errors import CrcError

from sml2mqtt import CONFIG


class StreamReaderGroup:
    def __init__(self, *crcs: str) -> None:
        self.readers = [SmlStreamReader(crc=crc) for crc in crcs]
        self.last_reader: SmlStreamReader | None = None

    def add(self, _bytes: bytes) -> None:
        for reader in self.readers:
            reader.add(_bytes)

    def get_frame(self) -> SmlFrame | None:
        crc_errors = []
        last_frame: SmlFrame | None = None
        for reader in self.readers:
            try:
                if (ret := reader.get_frame()) is None:
                    continue

                if last_frame is not None:
                    msg = 'Multiple frames'
                    raise ValueError(msg)

                if self.last_reader is not None and self.last_reader is not reader:
                    msg = 'Reader changed'
                    raise ValueError(msg)

                self.last_reader = reader
                last_frame = ret
            except CrcError as e:
                crc_errors.append(e)

        if last_frame is not None:
            return last_frame

        # all readers issued a crc error -> propagate
        if len(crc_errors) == len(self.readers):
            raise crc_errors[0] from None

        return None

    def get_reader(self) -> SmlStreamReader:
        if self.last_reader is None:
            msg = 'Last reader not set'
            raise ValueError(msg)
        return self.last_reader


def create_stream_reader_group() -> StreamReaderGroup:
    crc_algorithms = CONFIG.general.crc
    return StreamReaderGroup(*crc_algorithms)



from .date_time_finder import DateTimeFinder, get_now
from .protocols import DeviceProto, SourceProto
from .sml_helpers import EnhancedSmlFrame, SmlFrameValues
from .task import DeviceTask, Task, create_task
from .time_series import DurationType, TimeSeries, get_duration



import logging
import sys
from datetime import date, datetime
from logging.handlers import RotatingFileHandler
from pathlib import Path

import sml2mqtt


log = logging.getLogger('sml')


def get_logger(suffix: str) -> logging.Logger:
    if not suffix:
        return log
    return log.getChild(suffix)


class MidnightRotatingFileHandler(RotatingFileHandler):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.last_check: date = datetime.now().date()

    def shouldRollover(self, record) -> int:
        if (date_now := datetime.now().date()) == self.last_check:
            return 0
        self.last_check = date_now
        return super().shouldRollover(record)


def setup_log() -> None:
    level = sml2mqtt.CONFIG.logging.set_log_level()

    # This is the longest logger name str
    chars = 0
    for device in sml2mqtt.CONFIG.inputs:
        # Name of the longest logger, should be the device status
        chars = max(len(get_logger(device.get_device_name()).getChild('status').name), chars)
    log_format = logging.Formatter('[{asctime:s}] [{name:' + str(chars) + 's}] {levelname:8s} | {message:s}', style='{')

    # File Handler
    file_path = sml2mqtt.CONFIG.logging.file
    log_to_stdout = file_path.lower() == 'stdout'
    if log_to_stdout:
        handler = logging.StreamHandler(sys.stdout)
    else:
        log_file = Path(file_path)
        if not log_file.is_absolute():
            log_file = sml2mqtt.CMD_ARGS.config.parent / log_file
            log_file.resolve()

        handler = MidnightRotatingFileHandler(
            str(log_file), maxBytes=1024 * 1024, backupCount=3, encoding='utf-8'
        )

    handler.setFormatter(log_format)
    handler.setLevel(level)

    # Bind handler to logger
    root_log = logging.getLogger()
    root_log.addHandler(handler)

    # If we analyze we print, too
    if sml2mqtt.CMD_ARGS.analyze and not log_to_stdout:
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(level)
        ch.setFormatter(log_format)
        root_log.addHandler(ch)

    log.info(f'Starting V{sml2mqtt.__version__}')



from collections.abc import Generator
from time import monotonic
from typing import Final

from sml2mqtt.const import SmlFrameValues
from sml2mqtt.mqtt import MqttObj
from sml2mqtt.sml_value.base import OperationContainerBase, SmlValueInfo


class SmlValue(OperationContainerBase):
    def __init__(self, obis: str, mqtt: MqttObj) -> None:
        super().__init__()

        self.obis: Final = obis
        self.mqtt: Final = mqtt

        self.last_publish: float = 0

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} obis={self.obis} at 0x{id(self):x}>'

    def process_frame(self, frame: SmlFrameValues):
        if (sml_value := frame.get_value(self.obis)) is None:
            return None

        info = SmlValueInfo(sml_value, frame, self.last_publish)
        value = sml_value.get_value()

        for op in self.operations:
            value = op.process_value(value, info)

        if value is None:
            return None

        self.mqtt.publish(value)
        self.last_publish = monotonic()
        return value

    def describe(self) -> Generator[str, None, None]:
        yield f'<SmlValue>'
        yield f'  obis : {self.obis:s}'
        yield f'  topic: {self.mqtt.topic:s}'
        yield f'  operations:'
        for op in self.operations:
            yield from op.describe(f'    ')
        yield ''



from .actions import HeartbeatActionOperation, RefreshActionOperation
from .date_time import DateTimeFinder, MaxValueOperation, MinValueOperation, VirtualMeterOperation
from .filter import (
    DeltaFilterOperation,
    OnChangeFilterOperation,
    RangeFilterOperation,
    SkipZeroMeterOperation,
    ThrottleFilterOperation,
)
from .math import FactorOperation, OffsetOperation, RoundOperation, RoundToMultipleOperation
from .operations import OrOperation, SequenceOperation
from .time_series import MaxOfIntervalOperation, MeanOfIntervalOperation, MinOfIntervalOperation
from .workarounds import NegativeOnEnergyMeterWorkaroundOperation



import dataclasses
from collections.abc import Callable, Generator
from typing import Any, Final

from sml2mqtt.__log__ import get_logger
from sml2mqtt.config import OptionalMqttPublishConfig
from sml2mqtt.mqtt import publish

from .errors import MqttConfigValuesMissingError, MqttTopicEmpty, TopicFragmentExpectedError


pub_func: Callable[[str, int | float | str, int, bool], Any] = publish


def publish_analyze(topic: str, value: int | float | str, qos: int, retain: bool) -> None:
    get_logger('mqtt.pub').info(f'{topic}: {value} (QOS: {qos}, retain: {retain})')


def patch_analyze() -> None:
    global pub_func

    pub_func = publish_analyze


@dataclasses.dataclass
class MqttCfg:
    topic_full: str | None = None
    topic_fragment: str | None = None
    qos: int | None = None
    retain: bool | None = None

    def set_config(self, config: OptionalMqttPublishConfig) -> None:
        self.topic_full = config.full_topic
        self.topic_fragment = config.topic
        self.qos = config.qos
        self.retain = config.retain


class MqttObj:
    def __init__(self, topic_fragment: str | None = None, qos: int | None = None, retain: bool | None = None) -> None:

        # Configured parts
        self.cfg = MqttCfg(topic_fragment=topic_fragment, qos=qos, retain=retain)

        # Effective config
        self.qos: int = 0
        self.retain: bool = False
        self.topic: str = ''

        self.parent: MqttObj | None = None
        self.children: list[MqttObj] = []

    def publish(self, value: str | int | float) -> None:
        pub_func(self.topic, value, self.qos, self.retain)

    def update(self) -> 'MqttObj':
        self._merge_values()
        for c in self.children:
            c.update()
        return self

    def _merge_values(self) -> 'MqttObj':

        # no parent -> just set the config
        if self.parent is None:
            assert self.cfg.topic_full is None
            self.topic = self.cfg.topic_fragment    # expect fragment only
            self.qos = self.cfg.qos
            self.retain = self.cfg.retain
            if self.topic is None or self.qos is None or self.retain is None:
                raise MqttConfigValuesMissingError()
            return self

        # effective topic
        if self.cfg.topic_full:
            self.topic = self.cfg.topic_full
        else:
            if not self.cfg.topic_fragment:
                raise TopicFragmentExpectedError()

            # The topmost topic may be empty
            parts = []
            if self.parent.topic:
                parts.append(self.parent.topic)
            if self.cfg.topic_fragment:
                parts.append(self.cfg.topic_fragment)
            self.topic = '/'.join(parts)

            if not self.topic:
                raise MqttTopicEmpty()

        # effective QOS
        self.qos = self.cfg.qos
        if self.qos is None:
            self.qos = self.parent.qos

        # effective retain
        self.retain = self.cfg.retain
        if self.retain is None:
            self.retain = self.parent.retain
        return self

    def set_topic(self, topic: str | None) -> 'MqttObj':
        self.cfg.topic_fragment = topic
        self.update()
        return self

    def set_config(self, cfg: OptionalMqttPublishConfig | None) -> 'MqttObj':
        if cfg is None:
            return self

        self.cfg.set_config(cfg)
        self.update()
        return self

    def create_child(self, topic_fragment: str | None = None, qos: int | None = None,
                     retain: bool | None = None) -> 'MqttObj':
        child = self.__class__(topic_fragment=topic_fragment, qos=qos, retain=retain)
        child.parent = self
        self.children.append(child)
        child.update()
        return child

    def iter_objs(self) -> Generator['MqttObj', Any, None]:
        yield self
        for child in self.children:
            yield from child.iter_objs()


BASE_TOPIC: Final = MqttObj()


def setup_base_topic(topic: str, qos: int, retain: bool) -> None:
    BASE_TOPIC.cfg.topic_fragment = topic
    BASE_TOPIC.cfg.qos = qos
    BASE_TOPIC.cfg.retain = retain
    BASE_TOPIC.update()


def check_for_duplicate_topics(obj: MqttObj) -> None:
    log = get_logger('mqtt')

    topics: set[str] = set()
    for o in obj.iter_objs():
        if (topic := o.topic) in topics:
            log.warning(f'Topic "{topic:s}" is already configured!')
        topics.add(topic)



from easyconfig import AppBaseModel, BaseModel, Field, create_app_config

from .device import SmlDeviceConfig, SmlValueConfig
from .inputs import SerialSourceSettings, SmlSourceSettingType
from .logging import LoggingSettings
from .mqtt import MqttConfig, OptionalMqttPublishConfig
from .types import LowerStr, ObisHex


class GeneralSettings(BaseModel):
    wh_in_kwh: bool = Field(
        True, description='Automatically convert Wh to kWh', alias='Wh in kWh'
    )
    republish_after: int = Field(
        120, description='Republish automatically after this time (if no other filter configured)',
        alias='republish after',
    )
    report_blank_energy_meters: bool = Field(
        False, description='Report blank energy meters (where the value is 0kwh)',
        alias='report blank energy meters', in_file=False
    )
    report_device_id: bool = Field(
        False, description='Report the device id even though it does never change',
        alias='report device id', in_file=False
    )
    device_id_obis: list[ObisHex] = Field(
        # 0100000009ff (1-0:0.0.9*255) : Geräteeinzelidentifikation
        # 0100600100ff (1-0:96.1.0*255): Produktionsnummer
        ['0100000009ff', '0100600100ff'],
        description='Additional OBIS fields for the serial number to configuration matching',
        alias='device id obis', in_file=False
    )
    crc: list[str] = Field(
        default=['x25', 'kermit'],
        description='Which crc algorithms are used to calculate the checksum of the smart meter',
        alias='crc', in_file=False
    )


class Settings(AppBaseModel):
    logging: LoggingSettings = Field(default_factory=LoggingSettings)
    mqtt: MqttConfig = Field(default_factory=MqttConfig)
    general: GeneralSettings = Field(default_factory=GeneralSettings)
    inputs: list[SmlSourceSettingType] = Field(default_factory=list)
    devices: dict[LowerStr, SmlDeviceConfig] = Field({}, description='Device configuration by ID or url',)


def default_config() -> Settings:
    # File defaults
    return Settings(
        inputs=[SerialSourceSettings(type='serial', url='COM1', timeout=6),
                SerialSourceSettings(type='serial', url='/dev/ttyS0', timeout=6), ],
        devices={
            'device_id_hex': SmlDeviceConfig(
                mqtt=OptionalMqttPublishConfig(topic='DEVICE_BASE_TOPIC'),
                status=OptionalMqttPublishConfig(topic='status'),
                skip={'00112233445566'},
                values=[
                    SmlValueConfig(
                        obis='00112233445566',
                        mqtt=OptionalMqttPublishConfig(topic='OBIS'),
                        operations=[
                            {'negative on energy meter status': True},
                            {'factor': 3}, {'offset': 100}, {'round': 2},
                            {'type': 'change filter'},
                            {'refresh action': 600}
                        ]
                    )
                ]
            )
        }
    )


CONFIG: Settings = create_app_config(Settings(), default_config)



import random
import string

from easyconfig import BaseModel
from pydantic import Field, StrictBool, field_validator, model_validator

from sml2mqtt.config.mqtt_tls import MqttTlsOptions
from sml2mqtt.config.types import MqttQosInt, MqttTopicStr, StrippedStr


class MqttDefaultPublishConfig(BaseModel):
    qos: MqttQosInt = Field(
        0, description='Default value for QoS if no other QoS value in the config entry is set')
    retain: StrictBool = Field(
        False, description='Default value for retain if no other retain value in the config entry is set')


class OptionalMqttPublishConfig(BaseModel):
    topic: MqttTopicStr | None = Field(
        None, description='Topic fragment for building this topic with the parent topic')
    full_topic: MqttTopicStr | None = Field(
        None, alias='full topic', description='Full topic - will ignore the parent topic parts')
    qos: MqttQosInt | None = Field(
        None, description='QoS for publishing this value (if set - otherwise use parent)')
    retain: StrictBool | None = Field(
        None, description='Retain for publishing this value (if set - otherwise use parent)')

    @field_validator('topic', 'full_topic')
    def validate_topic(cls, value):
        if value is None:
            return None

        if value.endswith('/'):
            msg = 'Topic must not end with "/"'
            raise ValueError(msg)
        if value.startswith('/'):
            msg = 'Topic must not start with "/"'
            raise ValueError(msg)
        return value

    @model_validator(mode='after')
    def check_full_or_partial(self):
        if self.topic is not None and self.full_topic is not None:
            msg = 'Topic and full_topic can not be used at the same time!'
            raise ValueError(msg)
        return self


class MqttConnection(BaseModel):
    identifier: StrippedStr = Field('sml2mqtt-' + ''.join(random.choices(string.ascii_letters, k=13)),)
    host: StrippedStr = 'localhost'
    port: int = Field(1883, ge=0)
    user: StrippedStr = ''
    password: StrippedStr = ''

    tls: MqttTlsOptions | None = Field(None)


class MqttConfig(BaseModel):
    connection: MqttConnection = Field(
        default_factory=MqttConnection)
    topic: StrippedStr = Field(
        'sml2mqtt', alias='topic prefix', description='Prefix for all topics. Set to empty string to disable')
    defaults: MqttDefaultPublishConfig = Field(
        default_factory=MqttDefaultPublishConfig)
    last_will: OptionalMqttPublishConfig = Field(
        default_factory=lambda: OptionalMqttPublishConfig(topic='status'), alias='last will')



from collections.abc import Generator

from typing_extensions import override

from sml2mqtt.sml_value.base import OperationContainerBase, SmlValueInfo, ValueOperationBase


class OrOperation(ValueOperationBase, OperationContainerBase):

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        ret: float | None = None
        for op in self.operations:
            if (call := op.process_value(value, info)) is not None and ret is None:
                ret = call

        return ret

    def __repr__(self) -> str:
        return f'<Or at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Or:'
        for o in self.operations:
            yield from o.describe(indent + '  ')


class SequenceOperation(ValueOperationBase, OperationContainerBase):
    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        for op in self.operations:
            value = op.process_value(value, info)
        return value

    def __repr__(self) -> str:
        return f'<Sequence at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Sequence:'
        for o in self.operations:
            yield from o.describe(indent + '  ')



from collections.abc import Generator
from math import ceil, floor
from typing import Final, Literal

from typing_extensions import override

from sml2mqtt.sml_value.base import SmlValueInfo, ValueOperationBase


class FactorOperation(ValueOperationBase):
    def __init__(self, factor: int | float) -> None:
        self.factor: Final = factor

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None
        return value * self.factor

    def __repr__(self) -> str:
        return f'<Factor: {self.factor} at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Factor: {self.factor}'


class OffsetOperation(ValueOperationBase):
    def __init__(self, offset: int | float) -> None:
        self.offset: Final = offset

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None
        return value + self.offset

    def __repr__(self) -> str:
        return f'<Offset: {self.offset} at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Offset: {self.offset}'


class RoundOperation(ValueOperationBase):
    def __init__(self, digits: int) -> None:
        self.digits: Final = digits if digits else None

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None

        return round(value, self.digits)

    def __repr__(self) -> str:
        return f'<Round: {self.digits if self.digits is not None else 0} at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Round: {self.digits if self.digits is not None else "integer"}'


class RoundToMultipleOperation(ValueOperationBase):
    # noinspection PyShadowingBuiltins
    def __init__(self, value: int, round: Literal['up', 'down', 'nearest']) -> None:   # noqa: A002
        self.multiple: Final = value

        self.round_up: Final = round == 'up'
        self.round_down: Final = round == 'down'

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None

        if self.round_up:
            return self.multiple * int(ceil(value / self.multiple))
        if self.round_down:
            return self.multiple * int(floor(value / self.multiple))

        multiple = self.multiple
        div, rest = divmod(value, multiple)
        div = int(div)

        if rest >= 0.5 * multiple:
            return (div + 1) * multiple
        return div * multiple

    def __mode_str(self) -> str:
        if self.round_up:
            return 'up'
        if self.round_down:
            return 'down'
        return 'nearest'

    def __repr__(self) -> str:
        return f'<RoundToMultiple: value={self.multiple} round={self.__mode_str()} at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Round To Multiple:'
        yield f'{indent:s}      value: {self.multiple}'
        yield f'{indent:s}      round: {self.__mode_str()}'



from __future__ import annotations

from typing import TYPE_CHECKING, Final

from typing_extensions import override


if TYPE_CHECKING:
    from logging import Logger

    from smllib.sml import SmlListEntry


class Sml2MqttException(Exception):
    pass


class Sml2MqttExceptionWithLog(Sml2MqttException):
    def log_msg(self, log: Logger):
        raise NotImplementedError()


class DeviceFailedError(Sml2MqttException):
    pass


# ------------------------------------------------------------------------------------
# Initial setup failed
# ------------------------------------------------------------------------------------
class InitialSetupFailedError(Sml2MqttException):
    pass


class DeviceSetupFailedError(InitialSetupFailedError):
    pass


class InitialMqttConnectionFailedError(InitialSetupFailedError):
    pass


# ------------------------------------------------------------------------------------
# Config mapping errors
# ------------------------------------------------------------------------------------
class Sml2MqttConfigMappingError(Sml2MqttException):
    pass


class ObisIdForConfigurationMappingNotFoundError(Sml2MqttExceptionWithLog):
    @override
    def log_msg(self, log: Logger):
        return None


# ------------------------------------------------------------------------------------
# Source Errors
# ------------------------------------------------------------------------------------

class HttpStatusError(Sml2MqttExceptionWithLog):
    def __init__(self, status: int) -> None:
        super().__init__()
        self.status: Final = status

    def __str__(self) -> str:
        return f'{self.__class__.__name__:s}: {self.status:d}'

    @override
    def log_msg(self, log: Logger) -> None:
        log.error(f'Received http status {self.status}')

    def __eq__(self, other):
        if isinstance(other, HttpStatusError):
            return self.status == other.status
        return NotImplemented


# ------------------------------------------------------------------------------------
# Value Processing Errors
# ------------------------------------------------------------------------------------
class UnprocessedObisValuesReceivedError(Sml2MqttExceptionWithLog):
    def __init__(self, *values: SmlListEntry) -> None:
        super().__init__()
        self.values: Final = values

    @override
    def log_msg(self, log: Logger) -> None:
        log.error(f'Unexpected obis id{"" if len(self.values) == 1 else "s"} received!')
        for value in self.values:
            for line in value.format_msg().splitlines():
                log.error(line)


class RequiredObisValueNotInFrameError(Sml2MqttExceptionWithLog):
    def __init__(self, *obis: str) -> None:
        super().__init__()
        self.obis: Final = obis

    @override
    def log_msg(self, log: Logger) -> None:
        log.error(f'Expected obis id{"" if len(self.obis) == 1 else "s"} missing in frame: {", ".join(self.obis)}!')



from easyconfig import BaseModel
from pydantic import Field

from .mqtt import OptionalMqttPublishConfig
from .operations import OperationsListType
from .types import ObisHex


class SmlValueConfig(BaseModel):
    obis: ObisHex = Field(description='Obis code for this value')
    mqtt: OptionalMqttPublishConfig | None = Field(
        default=None,
        description='Mqtt config for this value (optional)'
    )

    operations: OperationsListType = Field(
        default=[],
        alias='operations',
        description='A sequence of operations that will be evaluated one after another.\n'
                    'If one operation blocks this will return nothing.'
    )


class SmlDeviceConfig(BaseModel):
    """Configuration for a sml device"""

    mqtt: OptionalMqttPublishConfig | None = Field(None, description='Optional MQTT configuration for this meter.')

    status: OptionalMqttPublishConfig = Field(
        default=OptionalMqttPublishConfig(topic='status'),
        description='Optional MQTT status topic configuration for this meter'
    )

    skip: set[ObisHex] = Field(
        default_factory=set, description='OBIS codes (HEX) of values that will not be published (optional)')

    values: list[SmlValueConfig] = Field(
        default=[], description='Configurations for each of the values (optional)')



from collections.abc import Generator, Sequence
from typing import Final

from typing_extensions import override

from sml2mqtt.const import TimeSeries
from sml2mqtt.sml_value.base import SmlValueInfo, ValueOperationBase
from sml2mqtt.sml_value.operations._helper import format_period


class TimeSeriesOperationBaseBase(ValueOperationBase):
    def __init__(self, time_series: TimeSeries, reset_after_value: bool) -> None:
        self.time_series: Final = time_series
        self.reset_after_value: Final = reset_after_value

        # Makes only sense in combination
        if reset_after_value and not time_series.wait_for_data:
            raise ValueError()

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}    Interval: {format_period(self.time_series.period)}'
        yield f'{indent:s}    Wait for data: {self.time_series.wait_for_data}'
        yield f'{indent:s}    Reset after value: {self.reset_after_value}'


class TimeSeriesOperationBase(TimeSeriesOperationBaseBase):
    def on_values(self, obj: Sequence[float]) -> float | None:
        raise NotImplementedError()

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:

        ts = info.frame.timestamp
        self.time_series.add_value(value, ts)

        if (values := self.time_series.get_values()) is None:
            return None

        if self.reset_after_value:
            self.time_series.clear()
        return self.on_values(values)


class TimeDurationSeriesOperationBase(TimeSeriesOperationBaseBase):
    def on_values(self, obj: Sequence[tuple[float, float]]) -> float | None:
        raise NotImplementedError()

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:

        ts = info.frame.timestamp
        self.time_series.add_value(value, ts)

        if (values := self.time_series.get_value_duration(ts)) is None:
            return None

        if self.reset_after_value:
            self.time_series.clear()
        return self.on_values(values)


class MaxOfIntervalOperation(TimeSeriesOperationBase):

    @override
    def on_values(self, obj: Sequence[float]) -> float | None:
        return max(obj)

    def __repr__(self) -> str:
        return f'<MaxOfInterval: interval={self.time_series.period}s at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Max Of Interval:'
        yield from super().describe(indent)


class MinOfIntervalOperation(TimeSeriesOperationBase):

    @override
    def on_values(self, obj: Sequence[float]) -> float | None:
        return min(obj)

    def __repr__(self) -> str:
        return f'<MinOfInterval: interval={self.time_series.period}s at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Min Of Interval:'
        yield from super().describe(indent)


class MeanOfIntervalOperation(TimeDurationSeriesOperationBase):

    @override
    def on_values(self, obj: Sequence[tuple[float, float]]) -> float | None:
        time = 0.0
        mean = 0.0
        for value, duration in obj:
            mean += value * duration
            time += duration

        if time <= 0:
            return None
        return mean / time

    def __repr__(self) -> str:
        return f'<MeanOfInterval: interval={self.time_series.period}s at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Mean Of Interval:'
        yield from super().describe(indent)



from __future__ import annotations

import logging.handlers
import signal
import traceback
from dataclasses import dataclass
from threading import Lock
from typing import TYPE_CHECKING, Final

from sml2mqtt.const import Task, create_task


if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable


log = logging.getLogger('sml.shutdown')


@dataclass(frozen=True)
class ShutdownObj:
    coro: Callable[[], Awaitable]
    msg: str

    async def do(self) -> None:
        try:
            log.debug(self.msg)
            await self.coro()
            log.debug(f'{self.msg:s} done!')
        except Exception as e:
            log.error(str(e))
            tb = traceback.format_exc().splitlines()
            for line in tb:
                log.error(line)


async def shutdown_coro() -> None:
    log.debug('Starting shutdown')
    for obj in SHUTDOWN_OBJS:
        await obj.do()
    log.debug('Shutdown complete')


def on_shutdown(coro: Callable[[], Awaitable], msg: str):
    global SHUTDOWN_OBJS
    for obj in SHUTDOWN_OBJS:
        if obj.coro == coro or obj.msg == msg:
            raise ValueError()
    SHUTDOWN_OBJS = (*SHUTDOWN_OBJS, ShutdownObj(coro, msg))


SHUTDOWN_OBJS: tuple[ShutdownObj, ...] = ()
SHUTDOWN_TASK: Final = Task(shutdown_coro, name='Shutdown Task')

SHUTDOWN_LOCK: Final = Lock()
SHUTDOWN_CALL: Task | None = None


async def do_shutdown_async() -> None:
    global SHUTDOWN_CALL

    try:
        if not SHUTDOWN_TASK.is_running:
            print('Shutting down ...')
            log.info('Shutting down ...')
            SHUTDOWN_TASK.start()
    finally:
        with SHUTDOWN_LOCK:
            SHUTDOWN_CALL = None


def do_shutdown():
    global SHUTDOWN_CALL

    with SHUTDOWN_LOCK:
        if SHUTDOWN_CALL is not None:
            return None
        SHUTDOWN_CALL = create_task(do_shutdown_async())


def _signal_handler_shutdown(sig, frame) -> None:
    do_shutdown()


def signal_handler_setup() -> None:
    signal.signal(signal.SIGINT, _signal_handler_shutdown)
    signal.signal(signal.SIGTERM, _signal_handler_shutdown)



from collections.abc import Callable
from dataclasses import dataclass
from inspect import getmembers, isclass
from pathlib import Path
from typing import Any

from _pytest.monkeypatch import derive_importpath
from easyconfig import yaml
from pydantic import BaseModel

import sml2mqtt
from sml2mqtt.config.inputs import SmlSourceSettingsBase
from sml2mqtt.config.operations import HasDateTimeFields, HasIntervalFields
from sml2mqtt.sml_value.setup_operations import MAPPING, setup_operations


@dataclass
class YamlBlock:
    file: Path
    line_no: int
    model: str
    lines: list[str]

    def validate(self, func: Callable[[BaseModel],Any] | None = None):
        sample_cfg = '\n'.join(self.lines)

        target, name = derive_importpath(self.model, True)
        model_cls = getattr(name, target)   # type: type[BaseModel]

        try:
            yaml_obj = yaml.yaml_rt.load(sample_cfg)
            model = model_cls.model_validate(yaml_obj)
            if func:
                func(model)
        except Exception:
            print('')
            print(f'Error in {self.file.parent.name}/{self.file.name}:{self.line_no}')
            raise


def validate_yaml_blocks(file: Path, prefix_model: str = 'yamlmodel: ', func: Callable[[BaseModel],Any] | None = None):

    current_module = ''

    model = ''
    obj: YamlBlock | None = None
    indentation = 0

    for line_no, line in enumerate(file.read_text().splitlines()):
        line = line
        line_low = line.strip().lower()

        if line_low.startswith('.. py:currentmodule::'):
            current_module = line.strip()[21:].strip()

        if line_low.startswith(prefix_model):
            assert not model
            model = line.strip()[len(prefix_model):]
            if '.' not in model:
                if not current_module:
                    msg = f'currentmodule not set in {file}:{line_no}'
                    raise ValueError(msg)
                model = f'{current_module}.{model}'

        if obj is not None:
            # we need one empty line before the yaml starts
            if not obj.lines and line_low:
                continue

            # find out indentation
            if not indentation and line_low:
                while line[indentation] == ' ':
                    indentation += 1

            # we have a non-indented line -> yaml is finished
            if line_low and line[0] != ' ':
                obj.validate(func)
                obj = None
                indentation = 0
                model = ''
                continue

            obj.lines.append(line)

        if line_low.startswith(('.. code-block:: yaml', '.. code-block:: yml')):
            if not model:
                msg = f'Model object not set in {file}:{line_no}'
                raise ValueError(msg)
            if obj is not None:
                msg = f'Object already set in {file}:{line_no}: {obj.model} line {obj.line_no}'
                raise ValueError(msg)

            obj = YamlBlock(file, line_no, model, [])

    if obj:
        obj.validate(func)


def test_yaml_samples(pytestconfig) -> None:

    class DummyOperationParent:
        def add_operation(self, obj) -> None:
            pass

    class HasOperationsModel(BaseModel):
        operations: list[BaseModel]

    def check_obj(model: BaseModel) -> None:
        if model.__class__ in MAPPING:
            setup_operations(DummyOperationParent(), HasOperationsModel(operations=[model]))

    for file in (pytestconfig.rootpath / 'docs').iterdir():
        if file.suffix.lower() == '.rst' and file.name.lower() != 'smart_home_systems.rst':
            validate_yaml_blocks(file, func=check_obj)


def _get_documented_objs(path: Path, objs: set[str]) -> None:

    current_module = ''

    for line in (x.strip().replace('  ', '') for x in path.read_text().splitlines()):  # type: str
        if line.startswith('.. py:currentmodule::'):
            current_module = line[21:].strip()
            continue

        if line.startswith('.. autopydantic_model::'):
            obj_name = line[23:].strip()
            if '.' not in obj_name:
                assert current_module
                obj_name = f'{current_module}.{obj_name}'
            assert obj_name not in objs
            objs.add(obj_name)


def test_config_documentation_complete(pytestconfig) -> None:
    cfg_model_dir: Path = pytestconfig.rootpath / 'src' / 'sml2mqtt' / 'config'
    assert cfg_model_dir.is_dir()

    documented_objs = set()
    _get_documented_objs(pytestconfig.rootpath / 'docs' / 'configuration.rst', documented_objs)
    _get_documented_objs(pytestconfig.rootpath / 'docs' / 'operations.rst', documented_objs)

    # get Config implementation from source
    existing_objs = set()
    for module_name in [f.stem for f in cfg_model_dir.glob('**/*.py')]:
        module = getattr(sml2mqtt.config, module_name)
        cfg_objs = [
            x[1] for x in getmembers(
                module, lambda x: isclass(x) and issubclass(x, BaseModel) and x not in (
                    SmlSourceSettingsBase, HasIntervalFields, HasDateTimeFields
                )
            )
        ]
        cfg_names = {
            f'{obj.__module__}.{obj.__qualname__}' for obj in cfg_objs if not obj.__module__.startswith('easyconfig.')
        }
        existing_objs.update(cfg_names)

        # we check this here to get the module with the error message
        missing = cfg_names - documented_objs
        assert not missing, module.__name__

    # ensure that everything that is implemented is documented
    assert existing_objs == documented_objs



from __future__ import annotations

from typing import TYPE_CHECKING, Final


if TYPE_CHECKING:
    from collections.abc import Generator

    from smllib.sml import SmlListEntry

    from sml2mqtt.const import SmlFrameValues


class SmlValueInfo:
    __slots__ = ('value', 'frame', 'last_pub')

    def __init__(self, sml: SmlListEntry, frame: SmlFrameValues, last_pub: float) -> None:
        self.value: Final = sml
        self.frame: Final = frame

        self.last_pub: Final = last_pub

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} obis={self.value.obis}>'


class ValueOperationBase:
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        raise NotImplementedError()

    def describe(self, indent: str = '') -> Generator[str, None, None]:
        raise NotImplementedError()


class OperationContainerBase:
    def __init__(self) -> None:
        self.operations: tuple[ValueOperationBase, ...] = ()

    def add_operation(self, operation: ValueOperationBase):
        self.operations = (*self.operations, operation)
        return self

    def insert_operation(self, operation: ValueOperationBase):
        self.operations = (operation, *self.operations)
        return self


class ValueOperationWithStartupBase(ValueOperationBase):
    _PROCESS_VALUE_BACKUP_ATTR: Final = '_process_value_original'

    def on_first_value(self, value: float, info: SmlValueInfo):
        raise NotImplementedError()

    def enable_on_first_value(self) -> None:
        name: Final = self._PROCESS_VALUE_BACKUP_ATTR
        if hasattr(self, name):
            raise ValueError()

        setattr(self, name, self.process_value)
        self.process_value = self._process_value_first

    def _process_value_first(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None

        # restore original function
        name: Final = self._PROCESS_VALUE_BACKUP_ATTR
        self.process_value = getattr(self, name)
        delattr(self, name)

        return self.on_first_value(value, info)



from collections.abc import Callable
from typing import Protocol

from pydantic import BaseModel

from sml2mqtt.config.operations import (
    DeltaFilter,
    Factor,
    HeartbeatAction,
    MaxOfInterval,
    MaxValue,
    MeanOfInterval,
    MinOfInterval,
    MinValue,
    NegativeOnEnergyMeterWorkaround,
    Offset,
    OnChangeFilter,
    OperationsType,
    Or,
    RangeFilter,
    RefreshAction,
    Round,
    RoundToMultiple,
    Sequence,
    ThrottleFilter,
    VirtualMeter,
)
from sml2mqtt.sml_value.base import OperationContainerBase, ValueOperationBase
from sml2mqtt.sml_value.operations import (
    DeltaFilterOperation,
    FactorOperation,
    HeartbeatActionOperation,
    MaxOfIntervalOperation,
    MaxValueOperation,
    MeanOfIntervalOperation,
    MinOfIntervalOperation,
    MinValueOperation,
    NegativeOnEnergyMeterWorkaroundOperation,
    OffsetOperation,
    OnChangeFilterOperation,
    OrOperation,
    RangeFilterOperation,
    RefreshActionOperation,
    RoundOperation,
    RoundToMultipleOperation,
    SequenceOperation,
    ThrottleFilterOperation,
    VirtualMeterOperation,
)


def create_workaround_negative_on_energy_meter(enabled_or_obis: bool | str):
    if isinstance(enabled_or_obis, str):
        return NegativeOnEnergyMeterWorkaroundOperation(meter_obis=enabled_or_obis)
    if enabled_or_obis:
        return NegativeOnEnergyMeterWorkaroundOperation()
    return None


def create_or(operations: list[OperationsType]):
    return OrOperation()


def create_sequence(operations: list[OperationsType]):
    return SequenceOperation()


MAPPING = {
    OnChangeFilter: OnChangeFilterOperation,
    HeartbeatAction: HeartbeatActionOperation,
    DeltaFilter: DeltaFilterOperation,
    RangeFilter: RangeFilterOperation,

    RefreshAction: RefreshActionOperation,
    ThrottleFilter: ThrottleFilterOperation,

    Factor: FactorOperation,
    Offset: OffsetOperation,
    Round: RoundOperation,
    RoundToMultiple: RoundToMultipleOperation,

    NegativeOnEnergyMeterWorkaround: create_workaround_negative_on_energy_meter,

    Or: create_or,
    Sequence: create_sequence,

    VirtualMeter: VirtualMeterOperation,
    MinValue: MinValueOperation,
    MaxValue: MaxValueOperation,

    MaxOfInterval: MaxOfIntervalOperation,
    MinOfInterval: MinOfIntervalOperation,
    MeanOfInterval: MeanOfIntervalOperation,
}


def get_operation_factory(obj: BaseModel) -> Callable:
    for cfg_type, func in MAPPING.items():
        if isinstance(obj, cfg_type):
            return func
    msg = f'Unknown operation configuration type {type(obj)}'
    raise ValueError(msg)


def get_kwargs_names(obj: BaseModel) -> list[str]:
    return [n for n in dir(obj) if n.startswith('get_kwargs_')]


class _HasOperationsProto(Protocol):
    operations: list[BaseModel]


def setup_operations(parent: OperationContainerBase, cfg_parent: _HasOperationsProto) -> None:
    for cfg in cfg_parent.operations:
        factory = get_operation_factory(cfg)

        if kwarg_names := get_kwargs_names(cfg):
            kwargs = {name: value for kwarg_name in kwarg_names for name, value in getattr(cfg, kwarg_name)().items()}
        else:
            kwargs = cfg.model_dump(exclude={'type'})

        if (operation_obj := factory(**kwargs)) is None:
            continue

        assert isinstance(operation_obj, ValueOperationBase)
        parent.add_operation(operation_obj)

        if isinstance(operation_obj, (OrOperation, SequenceOperation)):
            setup_operations(operation_obj, cfg)



from __future__ import annotations

from typing_extensions import Protocol


class DeviceProto(Protocol):
    @property
    def name(self) -> str:
        ...

    def on_source_data(self, data: bytes) -> None:
        ...

    def on_source_failed(self, reason: str) -> None:
        ...

    def on_error(self, e: Exception, *, show_traceback: bool = True) -> None:
        ...


class SourceProto(Protocol):
    def start(self) -> None:
        ...

    def cancel_and_wait(self) -> None:
        ...



from typing import Annotated, Literal, TypeAlias

from pydantic import Strict, StrictFloat, StrictInt, StringConstraints

from sml2mqtt.__log__ import get_logger


log = get_logger('config')


ObisHex = Annotated[
    str,
    StringConstraints(to_lower=True, strip_whitespace=True, pattern=r'[0-9a-fA-F]{12}', strict=True)
]

LowerStr = Annotated[
    str,
    StringConstraints(to_lower=True, strip_whitespace=True, strict=True)
]


Number: TypeAlias = StrictInt | StrictFloat

PercentStr = Annotated[str, Strict(), StringConstraints(strip_whitespace=True, pattern=r'^\d+\.?\d*\s*%$')]

StrippedStr = Annotated[str, Strict(), StringConstraints(strip_whitespace=True)]

MqttTopicStr = Annotated[str, Strict(), StringConstraints(strip_whitespace=True, min_length=1)]
MqttQosInt: TypeAlias = Literal[0, 1, 2]
